"""Provides functions to extract portions of the expression tree and
to construct new expressions.  Heavily used in eval.py"""

__license__ = "MIT License"

import sys
from sets import Set

import pair
import types
from symbol import isSymbol, false, Symbol
import unittest
import pogo
import parser


def isSelfEvaluating(exp):
    """Returns True if the expression is self-evaluating."""
    return isNumber(exp) or isString(exp) or pair.isNull(exp)


def isNumber(x):
    """Returns True if we see a number."""
    return type(x) in (types.IntType, types.FloatType, types.LongType)


def isString(x):
    """Returns True if we see a string."""
    return type(x) == types.StringType



def isVariable(exp):
    """Returns true if the expression looks like a symbol."""
    return isSymbol(exp)


def isTaggedList(exp, tag):
    """Returns true if the expression is tagged with the 'tag'."""
    if pair.isList(exp) and pair.schemeLength(exp) > 0:
        return pair.schemeCar(exp) == tag
    return 0


######################################################################

def isQuoted(exp):
    return (isTaggedList(exp, Symbol("quote")) or
            isTaggedList(exp, Symbol("quasiquote")))

def makeQuoted(exp):
    return pair.schemeList(Symbol("quote"), exp)


def isQuasiquoted(exp):
    return isTaggedList(exp, Symbol("quasiquote"))


def isUnquoted(exp):
    return isTaggedList(exp, Symbol("unquote"))


def textOfQuotation(exp):
    return pair.schemeCadr(exp)


def textOfUnquotation(exp):
    return pair.schemeCadr(exp)


def expandQuasiquotation(exp, depth=1):
    """Takes a quasiquoted expression and constructs a new quoted
    expression.

    FIXME: this function is SO evil.  *grin*  Must clean this up.
    """
    if not pair.isList(exp):
        return makeQuoted(exp)
    if isUnquoted(exp):
        if depth == 1:
            return textOfUnquotation(exp)
        else:
            return pair.schemeList(
                Symbol("list"),
                makeQuoted(Symbol("unquote")),
                expandQuasiquotation(textOfUnquotation(exp), depth - 1))
    if isQuasiquoted(exp):
        return pair.schemeList(
            Symbol("list"),
            makeQuoted(Symbol("quasiquote")),
            expandQuasiquotation(textOfUnquotation(exp), depth + 1))
    else:
        return pair.schemeCons(
            Symbol("list"),
            pair.listMap(lambda subexp:
                         expandQuasiquotation(subexp, depth),
                         exp))



######################################################################

def isLet(exp):
    return isTaggedList(exp, Symbol("let"))


def letBindings(exp):
    return pair.schemeCadr(exp)


def letBindingVariables(bindings):
    if pair.isNull(bindings): return pair.schemeList()
    return pair.schemeCons(pair.schemeCar(pair.schemeCar(bindings)),
                           letBindingVariables(pair.schemeCdr(bindings)))

def letBindingValues(bindings):
    if pair.isNull(bindings): return pair.schemeList()
    return pair.schemeCons(pair.schemeCadr(pair.schemeCar(bindings)),
                           letBindingValues(pair.schemeCdr(bindings)))

def letBody(exp):
    return pair.schemeCdr(pair.schemeCdr(exp))

def letToApplication(exp):
    bindings = letBindings(exp)
    return pair.schemeCons(makeLambda(letBindingVariables(bindings),
                                      letBody(exp)),
                      letBindingValues(bindings))

######################################################################


def isAssignment(exp):
    """Assignments have the form (set! <var> <value>)."""
    return isTaggedList(exp, Symbol("set!"))


def assignmentVariable(exp):
    return pair.schemeCadr(exp)


def assignmentValue(exp):
    return pair.schemeCaddr(exp)


######################################################################


def isDefinition(exp):
    return isTaggedList(exp, Symbol("define"))


def definitionVariable(exp):
    if isSymbol(pair.schemeCadr(exp)):
        return pair.schemeCadr(exp)
    return pair.schemeCar(pair.schemeCadr(exp))


def definitionValue(exp):
    if isSymbol(pair.schemeCadr(exp)):
        return pair.schemeCaddr(exp)
    return makeLambda(pair.schemeCdr(pair.schemeCadr(exp)),
                      pair.schemeCddr(exp))


######################################################################

def isLambda(exp):
    return isTaggedList(exp, Symbol("lambda"))


def lambdaParameters(exp):
    return pair.schemeCadr(exp)


def lambdaBody(exp):
    return pair.schemeCddr(exp)


def makeLambda(parameters, body):
    if pair.schemeCount(Symbol("."), parameters) > 1:
        raise SchemeError, \
              "Parameter list can have at most one vararg parameter -- LAMBDA"
    return pair.schemeAppend(pair.schemeList(Symbol("lambda"), parameters),
                             body)


######################################################################


def isIf(exp):
    return isTaggedList(exp, Symbol("if"))


def ifPredicate(exp):
    return pair.schemeCadr(exp)


def ifConsequent(exp):
    return pair.schemeCaddr(exp)


def ifAlternative(exp):
    if not pair.isNull(pair.schemeCdddr(exp)):
        return pair.schemeCadddr(exp)
    return false


def makeIf(predicate, consequent, alternative):
    return pair.schemeList(Symbol("if"), predicate, consequent, alternative)


######################################################################

def isBegin(exp):
    return isTaggedList(exp, Symbol("begin"))


def beginActions(exp):
    return pair.schemeCdr(exp)


def isLastExp(seq):
    return pair.isNull(pair.schemeCdr(seq))


def firstExp(seq):
    return pair.schemeCar(seq)


def restExps(seq):
    return pair.schemeCdr(seq)


def sequenceToExp(seq):
    if pair.isNull(seq): return seq
    if isLastExp(seq): return firstExp(seq)
    return makeBegin(seq)


def makeBegin(seq):
    return pair.schemeCons(Symbol("begin"), seq)


######################################################################


def isApplication(exp):
    return pair.isList(exp)


def operator(exp):
    if pair.schemeLength(exp) == 0:
        raise SchemeError, \
              "No operator given for procedure application -- OPERATOR"
    return pair.schemeCar(exp)


def operands(exp):
    return pair.schemeCdr(exp)


def isNoOperands(ops):
    return pair.isNull(ops)


def firstOperand(ops):
    return pair.schemeCar(ops)


def restOperands(ops):
    return pair.schemeCdr(ops)


######################################################################


def isCond(exp):
    return isTaggedList(exp, Symbol("cond"))


def condClauses(exp):
    return pair.schemeCdr(exp)


def isCondElseClause(clause):
    return condPredicate(clause) == Symbol("else")


def condPredicate(clause):
    return pair.schemeCar(clause)


def condActions(clause):
    return pair.schemeCdr(clause)


def condToIf(exp):
    return expandClauses(condClauses(exp))


def expandClauses(clauses):
    if pair.isNull(clauses):
        return false
    first = pair.schemeCar(clauses)
    rest = pair.schemeCdr(clauses)
    if isCondElseClause(first):
        if pair.isNull(rest):
            return sequenceToExp(condActions(first))
        raise SchemeError, "else clause isn't last -- condToIf" + clauses
    return makeIf(condPredicate(first),
                  sequenceToExp(condActions(first)),
                  expandClauses(rest))

def isTrue(x):
    return (x != false)


def isFalse(x):
    return (x == false)

                                
######################################################################


def makeProcedure(parameters, body, env):
    return pair.schemeList(Symbol("procedure"), parameters, body, env)

def isCompoundProcedure(p):
    return isTaggedList(p, Symbol("procedure"))

def procedureParameters(p):
    return pair.schemeCadr(p)

def procedureBody(p):
    return pair.schemeCaddr(p)

def procedureEnvironment(p):
    return pair.schemeCadddr(p)



######################################################################
"""Here's where we define the primitives of Scheme."""
def isPrimitiveProcedure(proc):
    return isTaggedList(proc, Symbol("primitive"))


def primitiveImplementation(proc):
    return pair.schemeCadr(proc)


def makePrimitiveProcedure(proc):
    return pair.schemeList(Symbol("primitive"), proc)



## We have to tag a continuation slightly different: applying a
## continuation abandons the current computation.
def isContinuationProcedure(proc):
    return isTaggedList(proc, Symbol("continuation"))


def continuationImplementation(proc):
    return pair.schemeCadr(proc)


def makeContinuationProcedure(proc):
    """Procedure is defined to be a function that takes a single 'val'
    argument."""
    return pair.schemeList(Symbol("continuation"), proc)


######################################################################

def expressionToString(expr, quoteStrings=1):
    """Given a Scheme expression, returns a string that tries to nicely
    return it as a string.
    
    If quoteStrings is true, puts quotes around string expressions.

    Notes: this is a little tricky just because we have to account for
    loopy structures.  We keep a set of expressions that we've already
    seen, to make sure we don't retrace any steps.
    """
    seenExpressionIds = Set()

    def isLoopyPair(expr):
        return (pair.isPair(expr) and
                id(pair.schemeCdr(expr)) in seenExpressionIds)

    def markExprAsSeen(expr):
        if not pair.isNull(expr):
            seenExpressionIds.add(id(expr))

    def t_expressionToString(expr, cont):
        """Helper function for expressionToString: written in CPS/trampolined
        form to avoid growing control context."""
        if id(expr) in seenExpressionIds:
            return pogo.apply_cont(cont, "[...]")

        if pair.isNull(expr):
            return pogo.apply_cont(cont, "()")
        if not pair.isPair(expr):
            return t_atomExpressionToString(expr, cont)
        elif isPrimitiveProcedure(expr) or isCompoundProcedure(expr):
            return t_procedureExpressionToString(expr, cont)
        else:
            return t_pairExpressionToString(expr, cont)

    def t_atomExpressionToString(expr, cont):
        """Converts an atomic expression (string, number) to a string.
        Written in CPS/trampolined form."""
        if type(expr) is str and quoteStrings:
            return pogo.apply_cont(cont, "\"%s\"" % expr.replace('"', '\\"'))
        else:
            return pogo.apply_cont(cont, str(expr))
        
    def t_procedureExpressionToString(expr, cont):
        """Converts a procedure expression to a string."""
        if isPrimitiveProcedure(expr):
            return pogo.apply_cont(cont,
                                   ('(%s %s <procedure-env>)') % (
                Symbol('primitive-procedure'),
                primitiveImplementation(expr)))
        elif isCompoundProcedure(expr):
            def c_parameters(parameterString):
                def c_body(bodyString):
                    return pogo.apply_cont(cont,
                                           '(%s %s %s <procedure-env>)' % 
                                           (Symbol('compound-procedure'),
                                            parameterString,
                                            bodyString))
                return t_expressionToString(procedureBody(expr), c_body)
            return t_expressionToString(procedureParameters(expr),
                                        c_parameters)
        
    def t_pairExpressionToString(expr, cont):
        """Converts a pair expression to a string."""
        pieces = []
        def loop(loop_exp, loop_cont):
            if pair.isNull(loop_exp):
                return pogo.apply_cont(loop_cont, pieces)
            elif pair.isDottedPair(loop_exp) or isLoopyPair(loop_exp):
                def c_car(carString):
                    def c_cdr(cdrString):
                        pieces.append(carString)
                        pieces.append(".")
                        pieces.append(cdrString)
                        return pogo.apply_cont(loop_cont, pieces)
                    return t_expressionToString(pair.schemeCdr(loop_exp), c_cdr)
                return t_expressionToString(pair.schemeCar(loop_exp), c_car)
            else:
                def c_car(carString):
                    pieces.append(carString)
                    return pogo.apply_cont(loop, pair.schemeCdr(loop_exp), loop_cont)
                return pogo.apply_cont(t_expressionToString,
                                       pair.schemeCar(loop_exp),
                                       c_car)

        markExprAsSeen(expr)
        return loop(expr,
                    lambda pieces: pogo.apply_cont(cont,
                                                   '(' +
                                                   ' '.join(pieces) +
                                                   ')'))

    return pogo.pogo_start(t_expressionToString, expr, pogo.land)



######################################################################


class ExpressionTests(unittest.TestCase):
    def setUp(self):
        self._oldrecursionlimit = sys.getrecursionlimit()
        sys.setrecursionlimit(50)

    def tearDown(self):
        sys.setrecursionlimit(self._oldrecursionlimit)

    
    def constructEvilNestedExpression(self, n):
        result = pair.schemeList()
        for i in xrange(n-1):
            result = pair.schemeList(result)
        return result


    def testSimpleExpressionToString(self):
        self.assertEquals("5", expressionToString(5))
        self.assertEquals("x", expressionToString(Symbol("x")))


    def testSimpleListExpressionToString(self):
        self.assertEquals("(x)", expressionToString(pair.schemeList(Symbol("x"))))
        self.assertEquals("(1 2)", expressionToString(pair.schemeList(1, 2)))
        self.assertEquals("(1 (2) 3)", expressionToString(
            pair.schemeList(1, pair.schemeList(2), 3)))
        
    def testDottedPairs(self):
        self.assertEquals("(1 . 2)", expressionToString(pair.schemeCons(1, 2)))


    def testNullList(self):
        self.assertEquals("()", expressionToString(pair.schemeList()))
        self.assertEquals("(() ())", expressionToString(
            pair.schemeList(pair.schemeList(), pair.schemeList())))


    def testTailRecursiveExpressionToString(self):
        n = 1000
        evilExpression = self.constructEvilNestedExpression(n)
        self.assertEquals('(' * n + ')' * n,
                          expressionToString(evilExpression))


    def testRecursiveExpressionsDontKillUs(self):
        ## Makes sure that circular structures do not kill us.
        loopyList = pair.schemeCons(1, 1)
        pair.schemeSetCdrBang(loopyList, loopyList)
        self.assertEquals("(1 . [...])", expressionToString(loopyList))


    def testLargerExpression(self):
        program = parser.parse("""
 (define (factorial x)
    (if (= x 0)
         1
         (* x (factorial (- x 1)))))""")
        self.assertEquals("(define (factorial x) (if (= x 0) 1 (* x (factorial (- x 1)))))",
                          expressionToString(program))


if __name__ == '__main__':
    unittest.main()
